
/*
	Copyright (C) 1994, UK, John M.Howells - All rights reserved

	Windows Time Zone Information - WIN_TZ

	Author: John M.Howells
	CIS ID: 100031,353

	The Windows NT time zone data is provided in terms, usually, of the
	'n'th day in a month, independent of the year. To make sure that the
	computer will behave as expected it is convenient to be able to display
	this as proper absolute (i.e. calendar) dates. This program does so,
	converting the Windows NT time zone data into something meaningful.
	
	For those time zones that do have a Winter/Summer change, the user has
	the choice of the << or >> buttons, or a Now! button. The display
	starts with the current year, and when one of the << or >> is clicked
	the time zone for the previous or next year, respectively, is displayed
	in the Dialog box. Each time a << or >> button is clicked the next year
	in the appropriate direction is displayed, the previous year if the <<
	button is used and the next year if the >> button is used. The Now!
	button takes the display back to the year for the current time.
	
	The program allows the years from 1901 to 2099 to be specified. For
	this range the leap years do not need special calculation. The program
	assumes that the Time Zone data is appropriate for all the years within
	this range, even though this is sometimes not really the case.

	There are two other cases:

	For the case where a time zone does not have a Winter/Summer change the
	program just displays a Dialog Box with the name of the time zone and
	the bias (offset) with respect to UTC. The <<, >> and Now! buttons are
	disabled.
	
	If the time zone data contains only absolute dates, and not the usual
	'Day-in-Month' data, the program displays a dialog box with those dates,
	but with the <<, >> and Now! buttons disabled.

	The program starts by displaying the Time Zone Information for the
	current year, and for the local computer Time Zone. If the program is
	running on Windows NT 3.5 or later (but *not* Windows NT 3.1 where the
	data is not in the Registry - it was held directly in various resources
	in the MAIN.CPL Control Panel applet) the program displays a drop-down
	list box of all the possible Time Zones known to the Registry, so that
	the user can select one. When one is selected the time is reset and the
	display changes to the current year. Also, running on Windows NT 3.5 the
	program displays the UTC (Universal Co-ordinated Time), the local time
	for the user, and the time in the selected Time Zone, based on the
	data from the Time Zones information from the Registry, and the times
	are displayed when the program is Minimized. So that the text is okay
	when copied to the clopboard and then pasted into Notepad (but not other
	applications such as Write or Word for Win 6) each line ends with \r\n
	rather than \n.

	Note that
	    UTC = local time + bias
	so that
	    local time = UTC - bias
	or that
	    bias = UTC - local time
	so that
	    a bias of +3 means that 12:00 UTC =  9:00 local time
	and
	    a bias of -3 means that 12:00 UTC = 15:00 local time

	Note that InvalidateRect is always called with the final parameter as
	NULL, indicating not to erase the background. This is done in order to
	avoid flicker in the window when redrawimg. Because the program uses
	wsprintf() and DrawText() to prepare the output this gives quite bad
	flicker if written directly to the display device context, so the
	program first writes to a compatible Memory Device Context, and then
	BitBlt()'s that to the screen. It would be possible to change the
	display to output each element separately, but most of this code 
	would still be needed for the copy to the Clipboard.
 
	==========  ==========  ==========  ==========  ==========  ==========

	Things to do:
		Display text according to locale ...
		Produce a decent output format ...
	but for something so simple ...

	Also, produce a program that allows the user to select a number of
	time zones and display the time in each of those zones simultaneously.

	Some of the code here is a little crude, particularly that for putting
	the text on the dialog box, but it works, so ...

	==========  ==========  ==========  ==========  ==========  ==========

	Modification History:

	Date	    Version	Modifier	Reason

	94-11-06    1.0		JMH		Initial version.
*/



#if	!defined( WIN32 ) && !defined( _WIN32 ) && !defined( __WIN32__ )

#error	Must be compiled for 32-bit operation only

#endif


#define	UNICODE				/* because the Time Zone names for Standard
					   (Winter) and Daylight Saving (Summer)
					   are Unicode it is simplest to do the
					   complete program as resolutely Unicode. */



/* you can uncomment *ONE* of the following to test for various specific
   values that the data can take. there is no need for test for no DST
   data available (i.e. standard time only) or day-in-month data as in
   Windows NT 3.5 these conditions are met by the normal time zones data */
/* #define	TEST_ABSOLUTE		/* to test for absolute dates only */
/* #define	TEST_NO_TZ_DATA		/* to test for No Time Zones data in Registry */



/* build-time flags */

#define	DISPLAY_3D		/* disable this define to remove all 3D display elements.
				   this is so that when Windows NT gets automagic 3D, as for
				   Windows 95, the program's own 3D can be disabled. but ... */
/* #define	TEXTBOX_3D	/* ... because of the way the main text window is drawn the
				   3D effect for it will not be preserved, so we still have
				   to do that here, though this is all theoretical so far */


#define	STRICT

#include <windows.h>
#include <windowsx.h>			/* for GetWindowInstance() */
#include <search.h>			/* for qsort() */

#include "win_tz.h"


#define	USER_OVERRIDE	LOCALE_NOUSEROVERRIDE
					/* change to 0 to allow the user to control the
					   date and time formats, though surprisingly if
					   you do so the CPU time consumed changes a lot!
					   like, an order of magnitude from 0.6% to 5% */

/* like this
#undef	USER_OVERRIDE
#define	USER_OVERRIDE	0
 /* */



#define	BOX_NAME	"Win32 Time Zone Information"

static	WCHAR	tzAppName[ ] = TEXT( APP_NAME );
static	WCHAR	tzCpyRite[ ] = TEXT( COPYRITE );
static	WCHAR	tzBoxName[ ] = TEXT( BOX_NAME );


/* bit set in GetVersion() when program is not running on Windows NT.
   in Win32s GetVersion() returns with only the top bit set, while in
   Windows 95 Beta [aka Windows 4.x, aka Chicago] GetVersion() returns
   [according to Pietrek, MSJ Extra (UK edition), October '94, Page 24]
   with the top two bits set. For this program the top bit suffices. */
#define	WINVER_NOT_WIN_NT		0x80000000


/* name of root key for Time Zones items - not present in
   Windows NT 3.1 unless someone has added it separately! */
#define	REG_TZS		L"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones"

#if	defined( TEST_NO_TZ_DATA )	/* if testing create an invalid key */
#undef	REG_TZS
#define	REG_TZS		L"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\NO Time Zones"
#endif					/* of testing for no Zime Zones data */

/* names of items in each sub-key in Time Zones data - sub-key is Standard name */
#define	REG_TZ_DLT	L"Dlt"		/* Daylight name for a zone */
#define	REG_TZ_TZI	L"TZI"		/* binary Time Zone Information data for a zone */
#define	REG_TZ_DISPLAY	L"Display"	/* Display data, such as place names, for a zone */


/* where to find the function in Windows NT 3.5 that converts the system
   time (UTC) to a time within a particular time zone. run-time linked so
   it does not get a failure on NT 3.1 where the function did not exist. */
#define	KERNEL_32	L"Kernel32.dll"				/* module to run-time link to */
#define	SYS_TO_TZ	"SystemTimeToTzSpecificLocalTime"	/* function in module to link to */


/* name of our help file */
#define	WIN_TZ_HLP	L"WIN_TZ.HLP"


/* timer stuff */
#define	TIM_ID_PERIODIC	0		/* ID of normal timer */
#define	TIM_ID_SPLASH	1		/* ID of Splash Screen delay timer */
#define	TIM_DEL_SPLASH	5000		/* milliseconds for Splash Screen */


/* colour definitions */
#define	DK_RED		RGB( 128,   0,   0 )	/* for splash screen text */
#define	DK_BLUE		RGB(   0,   0, 128 )	/* for About box text */


/* create a language id */
#define	LANG_USED	MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT )


/* size of an array in elements */
#define		SIZEOF( xx ) 	( sizeof( xx ) / sizeof( xx[0] ) )


/* strings for the "<n>th" 'weekday' in month format, leave
   plenty of space for text in languages other than English */
static	WCHAR	ordinals[ 5 ][ 32 ] = {
			L"1st",
			L"2nd",
			L"3rd",
			L"4th",
			L"5th (last)" };


/* origin for day of week and months are different */
#define	FIRST_DAY	0
#define	FIRST_MONTH	1

/* names of days - note the locale order is different from the
   day number order - filled in by calls to Windows NT, leave
   plenty of space for names in languages other than English */
static	struct
{
	DWORD	locale_sday;		/* the locale identifier for name */
	WCHAR	name[ 32 ];		/* space for name */
} days[ FIRST_DAY + 7 ] = {
/* Sunday */	{ LOCALE_SDAYNAME7 | USER_OVERRIDE , L"" },
		{ LOCALE_SDAYNAME1 | USER_OVERRIDE , L"" },
		{ LOCALE_SDAYNAME2 | USER_OVERRIDE , L"" },
		{ LOCALE_SDAYNAME3 | USER_OVERRIDE , L"" },
		{ LOCALE_SDAYNAME4 | USER_OVERRIDE , L"" },
		{ LOCALE_SDAYNAME5 | USER_OVERRIDE , L"" },
/* Saturday */	{ LOCALE_SDAYNAME6 | USER_OVERRIDE , L"" } };

/* names of months - filled in by calls to Windows NT, leave
   plenty of space for names in languages other than English */
static	struct
{
	DWORD	locale_smonth;		/* the locale identifier for name */
	WCHAR	name[ 32 ];		/* space for name */
	INT	days;			/* how many days in month */
} months[ FIRST_MONTH + 12 ] = {
		{ 0     			              , L"", 0  },
/* Jan */  	{ LOCALE_SMONTHNAME1  | USER_OVERRIDE , L"", 31 },
		{ LOCALE_SMONTHNAME2  | USER_OVERRIDE , L"", 28 },
		{ LOCALE_SMONTHNAME3  | USER_OVERRIDE , L"", 31 },
		{ LOCALE_SMONTHNAME4  | USER_OVERRIDE , L"", 30 },
		{ LOCALE_SMONTHNAME5  | USER_OVERRIDE , L"", 31 },
		{ LOCALE_SMONTHNAME6  | USER_OVERRIDE , L"", 30 },
		{ LOCALE_SMONTHNAME7  | USER_OVERRIDE , L"", 31 },
		{ LOCALE_SMONTHNAME8  | USER_OVERRIDE , L"", 31 },
		{ LOCALE_SMONTHNAME9  | USER_OVERRIDE , L"", 30 },
		{ LOCALE_SMONTHNAME10 | USER_OVERRIDE , L"", 31 },
		{ LOCALE_SMONTHNAME11 | USER_OVERRIDE , L"", 30 },
/* Dec */	{ LOCALE_SMONTHNAME12 | USER_OVERRIDE , L"", 31 } };

/* and character for time separator - filled in by call to Windows NT.
   system allows two characters for separator, so we err on generous side */
static	struct
{
	DWORD	locale_stime;		/* the locale identifier for time separator */
	WCHAR	string[ 8 ];		/* space for time separator string */
} time_sep = { LOCALE_STIME | USER_OVERRIDE , L"" };


/* global items */
static	TIME_ZONE_INFORMATION	tz_info;	/* time zone information data */
static	DWORD			tz_isdst;	/* time zone data status */
static	WCHAR			tz_disp[ 256 ];	/* time zone Display data */
static	TIME_ZONE_INFORMATION	tz_local;	/* time zone information for local time zone */

static	BOOL			val_local;	/* data is for local time zone */
static	BOOL			val_button;	/* year select buttons enabled/disabled */
static	BOOL			val_tzs;	/* time zones available */

static	HBRUSH			win_back_brush;	/* handle for the window background brush */
static	HBRUSH			txt_back_brush;	/* handle for text background brush */

#if	defined( DISPLAY_3D )

static	HPEN			wd_pen_light;	/* handle for wide pen for main window highlight */
static	HPEN			wd_pen_shadow;	/* handle for wide pen for main window shadow */

#endif

#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

static	HPEN			nrw_pen_light;	/* handle for a single width pen for highlight */
static	HPEN			nrw_pen_shadow;	/* handle for a single width pen for shadow */

#endif


/* function prototypes */
LRESULT CALLBACK WndProc( HWND, UINT, WPARAM, LPARAM );
/* the Borland compiler does not like it if the next two functions are
   declared as LRESULT, because the type expected by the last parameter
   of the DialogBox function is a function returning BOOL, and BOOL is
   typedef'ed to INT, whereas LRESULT is typedef'ed to LONG. BOOL is a
   bit silly, as that should be TRUE or FALSE, but the WM_CTRLCLRSTATIC
   response is a handle to a brush, so INT is the best compromise. */
INT	CALLBACK AboutDlgProc( HWND, UINT, WPARAM, LPARAM );
INT	CALLBACK SplashDlgProc( HWND, UINT, WPARAM, LPARAM );

static	BOOL	get_locale_strings( VOID );
static	BOOL	init_drop_down( HWND );
static	VOID	setup_time_zone( HWND, HWND, INT );
static	VOID	draw_buff( HWND, HDC, RECT, RECT, INT );
static	BOOL	get_buff( PWCHAR, HWND, INT );
static	LONG	get_tz_buff( PWCHAR, INT, PWCHAR, PWCHAR, LONG, LPSYSTEMTIME, LONG, LONG );
static	INT	Zeller( INT, INT, INT );
static	INT	relative_time( LPSYSTEMTIME );
static	PWCHAR	tz_times( HWND, PWCHAR, BOOL );
static	VOID	to_clip( HWND, INT );


/* the function that kicks off the program */
int WINAPI WinMain( HINSTANCE hInstance,
		    HINSTANCE hPrevInstance,
		    LPSTR     lpszCmdLine,
		    int	      nCmdShow )
{
	MSG		tz_msg;
	WNDCLASS	tz_class;
	HACCEL		tz_accel;
	HWND		tz_wnd;

	UNREFERENCED_PARAMETER( hPrevInstance );
	UNREFERENCED_PARAMETER( lpszCmdLine );
	UNREFERENCED_PARAMETER( nCmdShow );

	/* make sure running on Windows NT proper */
	if( ( GetVersion() & WINVER_NOT_WIN_NT ) != 0 )
	{
		/* the Visual C++ help file says that the MessageBox
		   function works with Win32s, but omits to say that
		   it does not work when UNICODE is enabled, so here
		   the explicit MessageBoxA (for ASCII) function is
		   used with unadorned strings. */
		MessageBoxA( HWND_DESKTOP, "Program only runs on Windows NT",
				BOX_NAME, MB_OK | MB_ICONSTOP );
		return( FALSE );
	}

	/* On Windows NT hPrevInstance is always NULL, so the program
	   looks elsewhere to find out if another copy is already running.
	   NULL for window name parameter because the caption contains
	   time strings when iconized. */
	if( NULL != ( tz_wnd = FindWindow( tzAppName, NULL ) ) )
	{
		/* only one invocation allowed - send a message to the
		   previous one to restore it and bring it to the front */
		SendMessage( tz_wnd, WM_SYSCOMMAND, SC_RESTORE, 0 );
		/* and quit */
		return( FALSE );
	}

	/* here create a specific brush to make sure the
	   program consistently uses the colour required */
	win_back_brush = CreateSolidBrush( GetSysColor( COLOR_BTNFACE ) );

#if	defined( DISPLAY_3D )

	/* two pens for main window surround */
	wd_pen_light = CreatePen( PS_SOLID, 2, GetSysColor( COLOR_BTNHIGHLIGHT ) );
	wd_pen_shadow = CreatePen( PS_SOLID, 2, GetSysColor( COLOR_BTNSHADOW ) );

#endif

#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

	/* special pens for text box surround */
	nrw_pen_light = CreatePen( PS_SOLID, 0, GetSysColor( COLOR_BTNHIGHLIGHT ) );
	nrw_pen_shadow = CreatePen( PS_SOLID, 0, GetSysColor( COLOR_BTNSHADOW ) );

	/* for text box in 3D create a background brush to match standard window background colour */
	txt_back_brush = CreateSolidBrush( GetSysColor( COLOR_WINDOW ) );

#else

	/* for no 3D text background same as window background */
	txt_back_brush = win_back_brush;

#endif

	/* register the window class - byte align for efficiency. */
	tz_class.style          = CS_BYTEALIGNWINDOW | CS_DBLCLKS;
	tz_class.lpfnWndProc    = WndProc;
	tz_class.cbClsExtra     = 0;
	tz_class.cbWndExtra     = DLGWINDOWEXTRA;
	tz_class.hInstance      = hInstance;
	tz_class.hIcon          = LoadIcon( hInstance, L"Icon" );
	tz_class.hCursor        = LoadCursor( NULL, IDC_ARROW );
	tz_class.hbrBackground  = win_back_brush;
	tz_class.lpszMenuName   = NULL;
	tz_class.lpszClassName  = tzAppName;

	RegisterClass( &tz_class );

	/* create a Dialog Box to use as the main window and show it */
	tz_wnd = CreateDialog( hInstance, L"DST_Dlg", 0, NULL );
	ShowWindow( tz_wnd, SW_SHOWDEFAULT );

	/* fill drop-down list with time zone names */
	if( FALSE == init_drop_down( tz_wnd ) )
	{
		MessageBoxEx( tz_wnd, L"Failed to initialize drop-down list",
				tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
		return( FALSE );
	}

	/* if program starts Iconized change caption to the various times */
	if( TRUE == IsIconic( tz_wnd ) )
	{
		tz_times( tz_wnd, NULL, FALSE );
	}

	/* set up keyboard accelerators */
	if( NULL == ( tz_accel = LoadAccelerators( hInstance, L"Keys" ) ) )
	{
		MessageBoxEx( tz_wnd, L"Failed to load Accelerators",
				tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
		return( FALSE );
	}

	/* start the Splash Screen */
	DialogBox( hInstance, L"SplashBox", tz_wnd, SplashDlgProc );

	/* set focus to Exit button after Splash Screen */
	SetFocus( GetDlgItem( tz_wnd, IDC_BUTTON_EXIT ) );

	/* and do slightly unusual keyboard translation and message handling.
	   the books and examples all have IsDialogMessage() called before
	   the TranslateAccelerator(), but here they are the other way round,
	   because both functions apply to the same Windows handle, in which
	   case if the functions are used in the more usual order then the
	   TranslateAccelerator() function will never get anything (!), as all
	   input will be eaten by the IsDialogMessage() function. By doing
	   the functions this way round TranslateAccelerator() does its thing
	   only for the Accelerated keys, but IsDialogMessage() is still able
	   to work for the Dialog keys, such as Tab, as the user would expect. */
	while( TRUE == GetMessage( &tz_msg, NULL, 0, 0 ) )
	{
		if( FALSE == TranslateAccelerator( tz_wnd, tz_accel, &tz_msg ) &&
		    FALSE == IsDialogMessage( tz_wnd, &tz_msg ) )
		{
			TranslateMessage( &tz_msg );
			DispatchMessage( &tz_msg );
		}
	}
	return( tz_msg.wParam );
}


/* function that responds to Windows messages */
LRESULT CALLBACK WndProc( HWND	 hwnd,
			  UINT	 message,
			  WPARAM wParam,
			  LPARAM lParam )
{
	static	HWND		text_box = NULL;	/* handle to text box */
	static	RECT		win_rect;		/* co-ordinates of main window */
	static	RECT		text_rect;		/* co-ordinates of text box */
	static	INT		yr_offset = 0;		/* offset from current year */
	static	HWND		focus_wnd = NULL;	/* to keep track of the focus */
	static	INT		sec_cnt = 0;		/* how many seconds prog running */
	static	HDC		mem_dc;			/* Memory DC for window */
	static	HGDIOBJ		orig_brush;		/* original brush from memory DC */
	static	HGDIOBJ		orig_bitmap;		/* original bitmap from memory DC */
	HBITMAP			mem_dc_bit;		/* bitmap put into memory DC */
	HDC			disp_dc;		/* DC for the display window */
	HMENU			sys_menu;		/* handle to system menu */

	switch( message )
	{
	case WM_CREATE:
		/* on creation the Dialog Box used for the main
		   window will be in the top left of the screen by
		   default, so here it is centred on the screen. */
		if( FALSE == GetWindowRect( hwnd, &win_rect ) ||
		    FALSE == MoveWindow( hwnd, ( GetSystemMetrics( SM_CXSCREEN ) - win_rect.right ) / 2,
						 ( GetSystemMetrics( SM_CYSCREEN ) - win_rect.bottom ) / 2,
						   win_rect.right, win_rect.bottom, FALSE ) )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd, L"Failed to centre the window",
				tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return( -1 );
		}

		/* get and check client rectangle co-ordinates for 3D display */
		if( FALSE == GetClientRect( hwnd, &win_rect ) )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd, L"Failed to get co-ordinates of main window",
				tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return( -1 );
		}

		/* window has no caption in dialog so set it here */			
		if( FALSE == SetWindowText( hwnd, tzBoxName ) )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd, L"Failed to set up Window caption",
				tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return( -1 );
		}

		/* get and check system menu handle and add Copy and About items to it */
		if( NULL == ( sys_menu = GetSystemMenu( hwnd, FALSE ) ) ||
		    FALSE == AppendMenu( sys_menu, MF_SEPARATOR, 0,		 NULL ) ||
		    FALSE == AppendMenu( sys_menu, MF_STRING,    IDC_MENU_COPY,	 L"Copy\tCtrl-C" ) ||
		    FALSE == AppendMenu( sys_menu, MF_SEPARATOR, 0,		 NULL ) ||
		    FALSE == AppendMenu( sys_menu, MF_STRING,    IDC_MENU_ABOUT, L"&About..." ) )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd, L"Failed to set up Menu",
				tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return( -1 );
		}

		/* get and check time zone information */
		if( TIME_ZONE_ID_INVALID == ( tz_isdst = GetTimeZoneInformation( &tz_info ) ) )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd,  L"Failed to get Time Zone Information",
					tzBoxName, MB_OK | MB_ICONEXCLAMATION, LANG_USED );
			return( -1 );
		}
		/* and save as local time zone information */
		tz_local = tz_info;

		/* fill and check the month and day names. it is easier to do this
		   now for all the names than to get the particular names required. */
		if( FALSE == get_locale_strings() )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd, L"Failed to get day name or month name or time separator",
					tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return( -1 );
		}

		/* get a display DC and create a compatible Memory DC. the
		   compatible Memory DC only has a 1-bit placeholder bit-map,
		   so here create a compatible bitmap the size of the whole
		   window, then select bitmap into the Memory DC context, and
		   set the background brush. the bitmap is a little larger
		   than needed, but handling is easier, and check as we go. */
 		if( NULL == ( disp_dc = GetDC( hwnd ) ) ||
		    NULL == ( mem_dc = CreateCompatibleDC( disp_dc ) ) ||
		    NULL == ( mem_dc_bit = CreateCompatibleBitmap( disp_dc, win_rect.right, win_rect.bottom ) ) ||
		    NULL == ( orig_bitmap = SelectObject( mem_dc, mem_dc_bit ) ) ||
		    NULL == ( orig_brush = SelectObject( mem_dc, win_back_brush ) ) ||
		    0 == ReleaseDC( hwnd, disp_dc ) )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd, L"Failed to create device context for output",
					tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return( -1 );
		}

		/* start the 1 second timer that drives the clock display */
		if( 0 == SetTimer( hwnd, TIM_ID_PERIODIC, 1000, NULL ) )
		{
			/* if failed report error and hara-kiri */
			MessageBoxEx( hwnd, L"Failed to start one second timer",
					tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return( -1 );
		}			

		return( 0 );

	case WM_ACTIVATE:
		/* by the time this happens the child controls exist,
		   so here get a handle to one of them, which is not
		   possible when the WM_CREATE message arrives. */
		if( NULL == text_box )
		{
			/* get handle to text control */
			text_box = GetDlgItem( hwnd, IDC_CONTROL_TEXT );
			/* get text window co-ordinates */
			GetClientRect( text_box, &text_rect );
			/* convert the co-ordinates of the text control so that
			   they are with respect to the main window. only have to
			   do this once, as position is relative to main window. */
			MapWindowPoints( text_box, hwnd, (LPPOINT)&text_rect, 2 );
		}
		return( 0 );

	case WM_ACTIVATEAPP:
		/* make sure that when reactivated the previous control still
		   has the focus. this works for the non-minimized case but
		   so far I can't get the minimzed case to do the same thing. */
		if( TRUE == wParam && NULL != focus_wnd )
		{
			/* program being activated and knows which control previously
			   had the focus so restore the focus to that control */
			SetFocus( focus_wnd );
		}
		else
		{
			/* if the program doesn't know which button has the focus
			   or is being deactivated remember which Window has the focus */
			focus_wnd = GetFocus();
		}
		return( 0 );

	case WM_CTLCOLORSTATIC:
		/* this message is for child controls - pick out the text box */
		if( text_box == (HWND)lParam )
		{
			/* if text box set the background to transparent as
			   it's only there to get its size and position. */
			return( (LRESULT)GetStockObject( HOLLOW_BRUSH ) );
		}
		/* otherwise do default processing */
		break;

	case WM_COMMAND:
		switch( LOWORD( wParam ) )
		{
		case IDC_BUTTON_EXIT:
		case IDC_KEY_EXIT:
			SendMessage( hwnd, WM_CLOSE, 0, 0 );
			return( 0 );

		case IDC_BUTTON_NEXT:
		case IDC_KEY_NEXT:
			/* display for next year */
			if( TRUE == val_button )
			{
				yr_offset++;
				InvalidateRect( hwnd, &text_rect, FALSE );
			}
			else
			{
				/* cannot change year if no DST for this time zone */
				MessageBeep( 0xFFFFFFFF );
			}
			return( 0 );

		case IDC_BUTTON_PREV:
		case IDC_KEY_PREV:
			/* display for prev year */
			if( TRUE == val_button )
			{
				yr_offset--;
				InvalidateRect( hwnd, &text_rect, FALSE );
			}
			else
			{
				/* cannot change year if no DST for this time zone */
				MessageBeep( 0xFFFFFFFF );
			}
			return( 0 );

		case IDC_BUTTON_NOW:
		case IDC_KEY_NOW:
			/* display for this year */
			if( TRUE == val_button )
			{
				yr_offset = 0;
				InvalidateRect( hwnd, &text_rect, FALSE );
			}
			else
			{
				/* cannot change year if no DST for this time zone */
				MessageBeep( 0xFFFFFFFF );
			}
			return( 0 );

		case IDC_BUTTON_HELP:
		case IDC_KEY_HELP:
			/* show help file */
			WinHelp( hwnd, WIN_TZ_HLP, HELP_CONTENTS, 0 );
			return( 0 );

		case IDC_KEY_ABOUT:
			/* About from keyboard */
			DialogBox( GetWindowInstance( hwnd ), L"AboutBox", hwnd, AboutDlgProc );
			return( 0 );

		case IDC_KEY_COPY:
			/* Copy (CTRL-C) from keyboard */
			to_clip( hwnd, yr_offset );
			return( 0 );

		case IDC_DROPDOWN_LIST:
			/* item invoked from the drop down list */
			switch( HIWORD( wParam ) )
			{
			case CBN_SELCHANGE:
				/* select information for currently selected Time Zone */
				setup_time_zone( hwnd, (HWND)lParam,
						SendMessage( (HWND)lParam, CB_GETCURSEL, 0, 0 ) );
				/* and display it */
				InvalidateRect( hwnd, &text_rect, FALSE );
			}
			return( 0 );
		}
		break;

	case WM_SYSCOMMAND:
		switch( wParam & 0xFFF0 )

		{
		case SC_RESTORE:
			/* blank out icon text to hide times
			   if Time Zones data is available */
			SetWindowText( hwnd, L"" );
			/* bring to foreground so that the window is visible
			   if a second program instance restores this one */
			SetForegroundWindow( hwnd );
			/* and show window */
			ShowWindow( hwnd, SW_RESTORE );
			/* reset caption text */
			SetWindowText( hwnd, tzBoxName );
			return( 0 );

		case SC_MINIMIZE:
			if( TRUE == val_tzs )
			{
				/* if the Time Zones data is available in the Registry
				   blank out caption text so it does not appear below icon */
                        	SetWindowText( hwnd, L"" );
			}
			/* minimize Window */
			ShowWindow( hwnd, SW_MINIMIZE );
			if( TRUE == val_tzs )
			{
				/* and if zone data available start icon caption text */
				tz_times( hwnd, NULL, FALSE );
			}
			return( 0 );

		case SC_MAXIMIZE:
			/* ignore a Maximize message */
			return( 0 );

		case IDC_MENU_COPY:
			/* Copy from system menu */
			to_clip( hwnd, yr_offset );
			return( 0 );

		case IDC_MENU_ABOUT:
			/* About from system menu */
			DialogBox( GetWindowInstance( hwnd ), L"AboutBox", hwnd, AboutDlgProc );
			return( 0 );
		}
		break;

	case WM_PAINT:
        	draw_buff( hwnd, mem_dc, win_rect, text_rect, yr_offset );
		return( 0 );

	case WM_TIMECHANGE:
		/* reset time zone data to local time zone */
		GetTimeZoneInformation( &tz_info );
		/* and save as local time zone information */
		tz_local = tz_info;
		/* drop through */

	case WM_WININICHANGE:
		/* for change to time separator string reset names of things - probably
		   only needed if user can over-ride locale information but do it anyway */
		get_locale_strings();
		/* create drop-down list again - eventually there may be a utility
		   that allows a new time zone to be created, or have one removed.
		   here it is *assumed* that the author of any such utility has the
		   nous to generate a time change message when a new zone is added
		   or one removed. here start the drop-down list again to bring the
		   list of time zones into step with whatever is now in the Registry,
		   maybe nothing if the user removes the Time Zones key. */
		init_drop_down( hwnd );

		/* create new background brush */
		win_back_brush = CreateSolidBrush( GetSysColor( COLOR_BTNFACE ) );
		/* set up brush for background */
		DeleteObject( (HGDIOBJ)SetClassLong( hwnd, GCL_HBRBACKGROUND, (LONG)win_back_brush ) );

#if	defined( DISPLAY_3D )

		/* for main window shading first tidy up */
		DeleteObject( wd_pen_light );
		DeleteObject( wd_pen_shadow );
		/* then create two pens for window surround */
		wd_pen_light = CreatePen( PS_SOLID, 2, GetSysColor( COLOR_BTNHIGHLIGHT ) );
		wd_pen_shadow = CreatePen( PS_SOLID, 2, GetSysColor( COLOR_BTNSHADOW ) );


#endif

#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

 		/* if 3D for text box track changes to shadow */
		DeleteObject( nrw_pen_light );
		DeleteObject( nrw_pen_shadow );
                /* special pens for text box surround */
		nrw_pen_light = CreatePen( PS_SOLID, 0, GetSysColor( COLOR_BTNHIGHLIGHT ) );
		nrw_pen_shadow = CreatePen( PS_SOLID, 0, GetSysColor( COLOR_BTNSHADOW ) );

		/* and track changes to window background colour - first delete old brush */
		DeleteObject( txt_back_brush );
		/* then create a background brush to match current window background colour */
		txt_back_brush = CreateSolidBrush( GetSysColor( COLOR_WINDOW ) );

#else

		/* if no 3D of any kind text background same as main Window */
		txt_back_brush = win_back_brush;

#endif

		/* finally update display - the one case where we erase the background */
		InvalidateRect( hwnd, NULL, TRUE );
		return( 0 );

	case WM_TIMER:
		sec_cnt++;
		/* first update appropriate display */
		if( TRUE == IsIconic( hwnd ) )
		{
			/* if program is icnonized update caption text.
			   on Windows NT 3.1 this will do nothing. */
			tz_times( hwnd, NULL, FALSE );
		}
		else
		{
			/* if the program is not iconized redraw the display. */
			InvalidateRect( hwnd, &text_rect, FALSE );
		}
		return( 0 );

	case WM_LBUTTONDBLCLK:
		/* this next bit is silly, but ...
		   well, if 'they' can have gang screens I am going to have a hidden
		   trick, though it won't be very hidden as the source code is being
		   distributed! here if you double click with with the left mouse
		   button while the CNTL and SHIFT keys are down, you get a display
		   of when the program was build, how long the program has been
		   running, the kernel and user times used by the process, and the
		   single CPU percentage occupancy of the process. */
		if( ( MK_CONTROL | MK_SHIFT ) == ( wParam & ( MK_CONTROL | MK_SHIFT ) ) )
		{
			WCHAR		tbuff[ 1024 ];
			FILETIME	creation_time;
			FILETIME	exit_time;
			FILETIME	kernel_time;
			FILETIME	user_time;
			SYSTEMTIME	sys_kernel_time;
			SYSTEMTIME	sys_user_time;
			INT		proc_time;

			if( FALSE == GetProcessTimes( GetCurrentProcess(),
						&creation_time, &exit_time, &kernel_time, &user_time ) ||
			    FALSE == FileTimeToSystemTime( &kernel_time, &sys_kernel_time ) ||
			    FALSE == FileTimeToSystemTime( &user_time, &sys_user_time ) )
			{
				wsprintf( tbuff, L"Failed to get Timings" );
			}
			else
			{
				/* here you get : as the time separator whatever the system is set for */
				proc_time = ( sys_kernel_time.wMilliseconds + sys_user_time.wMilliseconds ) +
					( sys_kernel_time.wSecond + sys_user_time.wSecond ) * 1000 +
					( sys_kernel_time.wMinute + sys_user_time.wMinute ) * 60 * 1000;
				proc_time *= 100;
				wsprintf( tbuff, L"Built\ton\t%S\r\n"
						 L"\tat\t%S\r\n"
						 L"\r\n"
						 L"Processor Usage Statistics\r\n"
						 L"\r\n"
						 L"Elapsed Time\t = %2d:%02d:%02d (approx)\r\n"
						 L"User Time   \t = %2d:%02d:%02d.%03d\r\n"
						 L"Kernel Time \t = %2d:%02d:%02d.%03d\r\n"
						 L"%% occupancy\t = %2d.%03d",
						 	__DATE__, __TIME__,
							sec_cnt / 3600, ( sec_cnt / 60 ) % 60, sec_cnt % 60,
							sys_user_time.wHour, sys_user_time.wMinute,
								sys_user_time.wSecond, sys_user_time.wMilliseconds,
							sys_kernel_time.wHour, sys_kernel_time.wMinute,
								sys_kernel_time.wSecond, sys_kernel_time.wMilliseconds,
							( proc_time / sec_cnt ) / 1000, ( proc_time / sec_cnt ) % 1000 );
			}
			MessageBox( hwnd, tbuff, tzBoxName, MB_OK );
		}
		return( 0 );

	case WM_CLOSE:
		/* tell WinHelp to close down our help file - if open */
		WinHelp( hwnd, WIN_TZ_HLP, HELP_QUIT, 0 );
		DestroyWindow( hwnd );
		return( 0 );

	case WM_DESTROY:
		/* start by killing the periodic timer */
		KillTimer( hwnd, TIM_ID_PERIODIC );

		/* because the program created the brush the program
		   has to destroy it after selecting original brush */
		DeleteObject( SelectObject( mem_dc, orig_brush ) );
		/* destroy created bitmap after selecting original */
		DeleteObject( SelectObject( mem_dc, orig_bitmap ) );
		/* and delete the compatible Memory DC */
		DeleteDC( mem_dc );

#if	defined( DISPLAY_3D )

		/* delete wide pens */
		DeleteObject( wd_pen_light );
		DeleteObject( wd_pen_shadow );

#endif

#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

		/* delete narrow pens */
		DeleteObject( nrw_pen_light );
		DeleteObject( nrw_pen_shadow );

		/* delete brush for window background */
		DeleteObject( txt_back_brush );

#endif

		PostQuitMessage( 0 );
		return( 0 );
	}
	return( DefWindowProc( hwnd, message, wParam, lParam ) );
}


/* the function for the About Dialog box.
   I would like to be able to Switch To... and then End Task
   when the program is processing the Dialog Box, but I can't
   see how. Maybe it's not possible, as NOTEPAD, CARDFILE and
   Word for Windows 6.0a all suffer the same problem when they
   have a Dialog box up, so it's not just my problem! */
INT CALLBACK AboutDlgProc( HWND   hDlg,
			   UINT   message,
			   WPARAM wParam,
			   LPARAM lParam )
{
	UNREFERENCED_PARAMETER( lParam );

	switch( message )
	{
	case WM_INITDIALOG:
		return( TRUE );

	case WM_CTLCOLORSTATIC:
		/* child Text controls - dark blue text */
		SetTextColor( (HDC)wParam, DK_BLUE );
		/* text background transparent */
		SetBkMode( (HDC)wParam, TRANSPARENT );
		/* and control background transparent */
		return( (INT)GetStockObject( HOLLOW_BRUSH ) );

	case WM_CTLCOLORDLG:
		/* set colour of Dialog background to light grey */
		return( (INT)GetStockObject( LTGRAY_BRUSH ) );

	case WM_COMMAND:
		switch( LOWORD( wParam ) )
		{
		case IDOK:
			/* kill Dialog when OK button pressed */
			EndDialog( hDlg, 0 );
			return( TRUE );
		}
	}
	return( FALSE );
}


/* the function for the start-up Splash Screen (aka Dialog box).
   if user clicks mouse kill the timer but keep Splash Screen. this
   is only so that I can check the appearance without it timing out
   as it stays on screen as long as the mouse button is held down.
   actually, if you click in the Spash Screen window and then drag
   outside it the Splash Screen will stay on display indefinitely,
   or until you release the mouse within the Splash Screen window. */
INT CALLBACK SplashDlgProc( HWND   hDlg,
			    UINT   message,
			    WPARAM wParam,
			    LPARAM lParam )
{
	UNREFERENCED_PARAMETER( lParam );

	switch( message )
	{
	case WM_INITDIALOG:
		/* kick off the timer that kills off the Splash Screen */
		SetTimer( hDlg, TIM_ID_SPLASH, TIM_DEL_SPLASH, NULL );
		return( TRUE );

	case WM_CTLCOLORSTATIC:
		/* child Text controls - dark red text */
		SetTextColor( (HDC)wParam, DK_RED );
		/* text background transparent */
		SetBkMode( (HDC)wParam, TRANSPARENT );
		/* and control background transparent */
		return( (INT)GetStockObject( HOLLOW_BRUSH ) );

	case WM_CTLCOLORDLG:
		/* set colour of Dialog background to light grey */
		return( (INT)GetStockObject( LTGRAY_BRUSH ) );

	case WM_TIMER:
		/* kill timer */
		KillTimer( hDlg, TIM_ID_SPLASH );
		/* and kill Splash Screen */
		EndDialog( hDlg, 0 );
		return( TRUE );

	case WM_LBUTTONDOWN:
		/* left button clicked within Splash Screen.
		   kill timer so it stays on screen */
		KillTimer( hDlg, TIM_ID_SPLASH );
		return( TRUE );

	case WM_LBUTTONUP:
		/* when left button released within Splash Screen kill it */
		EndDialog( hDlg, 0 );
		return( TRUE );
	}
	return( FALSE );
}


/* function to fill month and day names.
   ideally the text for 1st, 2nd, 3rd, etc., and for Summer/Winter should be
   locale sensitive, and extracted here from String resources, but not yet! */
static	BOOL	get_locale_strings( VOID )
{
	INT	cnt;

	/* read day names */
	for( cnt = FIRST_DAY ; cnt < SIZEOF( days ) ; cnt++ )
	{
		if( 0 == GetLocaleInfoW( MAKELCID( LANG_USED, SORT_DEFAULT ), days[ cnt ].locale_sday,
			days[ cnt ].name, SIZEOF( days[0].name ) ) )
		{
			return( FALSE );
		}
	}
	/* and read month names */
	for( cnt = FIRST_MONTH ; cnt < SIZEOF( months ) ; cnt++ )
	{
		if( 0 == GetLocaleInfoW( MAKELCID( LANG_USED, SORT_DEFAULT ), months[ cnt ].locale_smonth,
			months[ cnt ].name, SIZEOF( months[0].name ) ) )
		{
			return( FALSE );
		}
	}
	/* and time separator */
	if( 0 == GetLocaleInfoW( MAKELCID( LANG_USED, SORT_DEFAULT ), time_sep.locale_stime,
			time_sep.string, SIZEOF( time_sep.string ) ) )
	{
		return( FALSE );
	}
	return( TRUE );
}


/* compare function for qsort() - in Windows NT 3.5 the names actually
   come out of the Registry in alphabetical order, so it's not obvious
   that the sort does anything, for a simple test swap the two strings
   in the lstrcmp() function and the names should come out in reverse */
static	int	_cdecl	qsort_lstrcmp( const void *str1, const void *str2 )
{
	return( lstrcmp( (PWCHAR)str1, (PWCHAR)str2 ) );
}


/* function to set up drop down list box from Time Zones data in the Registry */
static	BOOL	init_drop_down( HWND hwnd )
{
	HKEY		key;			/* Registry key for Time Zones */
	DWORD		key_ix;			/* index through sub-keys */
	HWND		drop_list;		/* handle to drop-down list */

	/* data for RegQueryInfoKey function - most not used here */
	WCHAR		class_string[ 256 ];	/* class name */
	DWORD	auto	class_string_siz = SIZEOF( class_string );
	DWORD		num_sub_keys;		/* number of subkeys */
	DWORD		longest_sub_key;	/* longest subkey name length */
	DWORD		longest_class_name_siz;	/* longest class string length */
	DWORD		num_values;		/* number of value entries */
	DWORD		longest_value_name_siz;	/* longest value name length */
	DWORD		longest_value_data_siz;	/* longest value data length */
	DWORD		sec_desc_siz;		/* security descriptor length */
	FILETIME	last_write_time;	/* last write time - also used by RegEnumKeyEx */

	/* ptr to memory allocated for list of names */
	struct	tz_key
	{
		WCHAR	tz_name[ SIZEOF( tz_info.StandardName ) ];
	} *tz_key_ptr;

	/* get handle to drop-down list */
	if( NULL == ( drop_list = GetDlgItem( hwnd, IDC_DROPDOWN_LIST ) ) )
	{
		return( FALSE );
	}

	/* try and open time zones key */
	if( ERROR_SUCCESS != RegOpenKeyEx( HKEY_LOCAL_MACHINE, REG_TZS, 0, KEY_READ, &key ) )
	{
		/* on Windows NT 3.1 the Time Zones data does not exist
		   in the registry, so kill off the drop down list box.	*/
		ShowWindow( drop_list, SW_HIDE );
		val_tzs = FALSE;
		/* note that failure does *NOT* count as an error */
		return( TRUE );
	}

	/* start with drop-down list empty */
	SendMessage( drop_list, CB_RESETCONTENT, 0, 0 );
	/* and create first entry as local time zone */
	SendMessage( drop_list, CB_ADDSTRING, 0, (LPARAM)L"[ Local Time Zone ]" );
	/* set that as the selection */
	SendMessage( drop_list, CB_SETCURSEL, 0, 0 );
	/* in local time zone */
	val_local = TRUE;

	/* this may seem slightly contorted, but there is no guarantee that the
	   sub-keys will be read in alphabetic order, so they are read and then
	   sorted. the simpler way would be something like:

	   	while( ERROR_SUCCESS != RegEnumKey( key, key_ix++, temp_str, ...... )
	   	{
	   		SendMessage( ...... (LPARAM)temp_str );
		}

	   iterating until RegEnumKey indicates that there are no more sub-keys.
	   the drop-down list could be set for an alphabetic order display, but
	   then the first item, the "[ Local Time Zone ]" string, might not be
	   in the right place, so we do it the slightly harder way.  */
	   
	/* find out information about the key - each time zone is one sub-key below this key */
	if( ERROR_SUCCESS != RegQueryInfoKey( key, class_string, &class_string_siz, NULL,
						&num_sub_keys, &longest_sub_key, &longest_class_name_siz,
						&num_values, &longest_value_name_siz, &longest_value_data_siz,
						&sec_desc_siz, &last_write_time ) ||
	    longest_sub_key >= SIZEOF( tz_info.StandardName ) ||
	    NULL == ( tz_key_ptr = (struct tz_key *)GlobalAlloc( GPTR, num_sub_keys * sizeof( struct tz_key ) ) ) )
	{
		/* either query failed, a key is too large, or
		   bad memory allocation for resulting strings */
		RegCloseKey( key );
		/* the memory cannot be allocated so no need to free it */
		return( FALSE );
	}

	/* we now know how many sub-keys there are, each of which
	   represents the name of a time zone, read them all in. */
	for( key_ix = 0 ; key_ix < num_sub_keys ; key_ix++ )
	{
		/* this must be auto as the function updates it with the number of characters
		   read and it needs to be set to the buffer size on function entry */
		DWORD	auto	size_tz_name = SIZEOF( tz_info.StandardName );

		if( ERROR_SUCCESS != RegEnumKeyEx( key, key_ix,
						( tz_key_ptr + key_ix )->tz_name, &size_tz_name,
						NULL, NULL, NULL, &last_write_time ) )
		{
			/* it should work - so if it fails ... */
			return( FALSE );
		}
	}

	/* sort the keys into alphabetic order and use them to fill drop-down list */
	qsort( tz_key_ptr, num_sub_keys, sizeof( struct tz_key ), qsort_lstrcmp );
	for( key_ix = 0 ; key_ix < num_sub_keys ; key_ix++ )
	{
		SendMessage( drop_list, CB_ADDSTRING, 0, (LPARAM)( ( tz_key_ptr + key_ix )->tz_name ) );
	}

	GlobalFree( tz_key_ptr );		/* free memory for time zone names */
	RegCloseKey( key );			/* free master key for time zone names */
	val_tzs = TRUE;				/* mark time zone names as valid */
	return( TRUE );
}


/* sets up the data for a user-selected time zone. the "MUST be auto"
   variables are (unusually!) declared as 'auto' and have to be auto
   because they are changed by the RegQueryValueEx function, and have
   to have the initial value reset each time this function is called.

   NOTE: that as of now (23rd September 1994) I have not seen a definition
   of the format of the Time Zones data (*NOT* the TimeZoneInformation
   which is documented) so this uses my deductions about the format.

   there is no need to check if Time Zones data is available in the
   Registry, as if it is not this is only ever called with ix of zero */
static	VOID	setup_time_zone( HWND hwnd,		/* main window handle */
				 HWND drop_list,	/* drop-down list handle */
				 INT  ix )		/* selection index in drop-down list */
{
	HKEY		mast_key;			/* master key for Time Zones */
	HKEY		sub_key;			/* sub-key for one Time Zone */
	DWORD		key_type;		/* to make sure the result is as expected */
	WCHAR		std_chs[ SIZEOF( tz_info.StandardName ) ];
						/* standard time zone name */
	WCHAR		dst_chs[ SIZEOF( tz_info.DaylightName ) ];
						/* daylight time zone name */
	DWORD	auto	siz_dst_chs = sizeof( dst_chs );
						/* in bytes - MUST be auto as it gets updated */
	struct
	{
		LONG		Bias;
		LONG		StandardBias;
		LONG		DaylightBias;
		SYSTEMTIME	StandardDate;
		SYSTEMTIME	DaylightDate;
	} tz_data;				/* deduced binary time zone data. */
	DWORD	auto	siz_tz_data = sizeof( tz_data );
						/* in bytes - MUST be auto as it gets updated */
	WCHAR		tz_disp_temp[ 256 ];	/* for Display */
	DWORD	auto	siz_tz_disp_temp = sizeof( tz_disp_temp );
						/* in bytes - MUST be auto as it gets updated */
	INT		tz_disp_cnt;

	if( 0 == ix )
	{
		/* go back to data for local time zone */
		tz_isdst = GetTimeZoneInformation( &tz_info );
		/* and save as local time zone information */
		tz_local = tz_info;
		/* selection is local time zone */
		val_local = TRUE;
	}
	else
	{
		/* open time zones master key */
		if( ERROR_SUCCESS != RegOpenKeyEx( HKEY_LOCAL_MACHINE, REG_TZS, 0, KEY_READ, &mast_key ) )
		{
			MessageBoxEx( hwnd, L"Failed to open Registry key for time zones",
					tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			return;
		}

		/* get name of selected time zone from drop-down list box - its sub-key */
		SendMessage( drop_list, CB_GETLBTEXT, ix, (LPARAM)std_chs );
		/* open that sub-key in time zones data */
		if( ERROR_SUCCESS != RegOpenKeyEx( mast_key, std_chs, 0, KEY_READ, &sub_key ) )
		{
			/* close the master key and report error */
			RegCloseKey( mast_key );
			MessageBoxEx( hwnd, L"Failed to open Registry sub-key for time zone",
				tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			/* mark data as suspect */
			tz_isdst = TIME_ZONE_ID_UNKNOWN;
			return;
		}
		/* read three items of data from time zone - the Daylight
		   name, the Time Zone Information, and the Display text,
		   and check that each data type is as expected. */
		if( ERROR_SUCCESS != RegQueryValueEx( sub_key, REG_TZ_DLT, 0,
						&key_type, (LPBYTE)dst_chs, &siz_dst_chs ) ||
		    REG_SZ != key_type ||
		    ERROR_SUCCESS != RegQueryValueEx( sub_key, REG_TZ_TZI, 0,
						&key_type, (LPBYTE)&tz_data, &siz_tz_data ) ||
		    REG_BINARY != key_type ||
		    ERROR_SUCCESS != RegQueryValueEx( sub_key, REG_TZ_DISPLAY, 0,
						&key_type, (LPBYTE)tz_disp_temp, &siz_tz_disp_temp ) ||
		    REG_SZ != key_type )
		{
			MessageBoxEx( hwnd, L"Failed to read key value",
					tzBoxName, MB_OK | MB_ICONSTOP, LANG_USED );
			/* mark data as suspect */
			tz_isdst = TIME_ZONE_ID_UNKNOWN;
			return;
		}
		/* tidy up keys */
		RegCloseKey( sub_key );
		RegCloseKey( mast_key );

		/* data is not local time zone */
		val_local = FALSE;

		/* terminate strings (in case not already terminated!) for
		   Standard and Daylight and copy to time zone structure */
		std_chs[ 31 ] = L'\0';
		lstrcpy( tz_info.StandardName, std_chs );
		dst_chs[ 31 ] = L'\0';
		lstrcpy( tz_info.DaylightName, dst_chs );

		/* copy date information */
		tz_info.DaylightDate = tz_data.DaylightDate;
		tz_info.StandardDate = tz_data.StandardDate;
		/* and the biases */
		tz_info.Bias         = tz_data.Bias;
		tz_info.StandardBias = tz_data.StandardBias;
		tz_info.DaylightBias = tz_data.DaylightBias;

		/* remove GMT+/- from start of Display (place) name(s) */
		tz_disp_cnt = 0;
		/* copy Display after (GMT+/-...) field */
		while( L'\0' != tz_disp_temp[ tz_disp_cnt ] )
		{
			if( L' ' == tz_disp_temp[ tz_disp_cnt ] )
			{
				/* found first space - copy from next character */
				lstrcpy( tz_disp, &tz_disp_temp[ tz_disp_cnt + 1 ] );
				return;
			}
			tz_disp_cnt++;
		}
		/* didn't find space - copy everything */
		lstrcpy( tz_disp, tz_disp_temp );
	}
}


#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

#define	COLOUR_TEXT	COLOR_WINDOWTEXT	/* for 3D text box, text colour as for Window text */

#else

#define	COLOUR_TEXT	COLOR_BTNTEXT		/* for no 3D text box, colour as for button face text */

#endif


/* function to draw DST data for one year. drawing is done
   indirectly via a compatible DC to minimize flicker. */
static	VOID	draw_buff( HWND  hwnd,		/* main window handle */
			   HDC   mem_dc,	/* memory DC */
			   RECT	 win_rect,	/* main window co-ordinates */
			   RECT	 text_rect,	/* area to write text to */
			   INT	 yr_offset)	/* year with respect to current year */
{
	BOOL		dst_ok;			/* whether DST data exists */
	WCHAR		obuff[ 1024 ];		/* output buffer */
	HDC		disp_dc;		/* DC for window */
	PAINTSTRUCT	paint_struct;
	TEXTMETRIC	text_met;
	RECT		text_out;

#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

	HGDIOBJ		orig_pen;		/* original pen used in window */

#endif

#if	defined( DISPLAY_3D )	

	/* if we are creating 3D we have to work out where drop-down list is displayed */

	static	HWND	dd_wnd = NULL;		/* handle to drop-down list */
	static	RECT	dd_rect;		/* relative co-ordinates of drop-down list */

	if( NULL == dd_wnd )
	{
		/* get handle to drop-down list control */
		dd_wnd = GetDlgItem( hwnd, IDC_DROPDOWN_LIST );
		/* get drop-down list co-ordinates */
		GetClientRect( dd_wnd, &dd_rect );
		/* convert the co-ordinates of the drop-down list control so
		   that they are with respect to the main window. only have to
		   do this once, as position is relative to main window. */
		MapWindowPoints( dd_wnd, hwnd, (LPPOINT)&dd_rect, 2 );
	}

#endif

	if( NULL == ( disp_dc = BeginPaint( hwnd, &paint_struct ) ) )
	{
		return;
	}

	/* get text metrics for height of characters */
	GetTextMetrics( disp_dc, &text_met );

	/* set background for text in main window to see-through */
	SetBkMode( disp_dc, TRANSPARENT );
	/* make Window text colour as for button face text */
	SetTextColor( disp_dc, GetSysColor( COLOR_BTNTEXT ) );
	/* and output Copyright text on main window */
	DrawText( disp_dc, tzCpyRite, -1, &text_rect, DT_CENTER );

	/* space main text down one and a quarter lines */
	text_rect.top += ( text_met.tmHeight * 5 ) / 4;

	/* erase here. zap whole memory DC background */
	FillRect( mem_dc, &text_rect, txt_back_brush );
	/* set text color to match configuration */
	SetTextColor( mem_dc, GetSysColor( COLOUR_TEXT ) );
	/* set background for text in memory DC to see-through */
	SetBkMode( mem_dc, TRANSPARENT );

#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

	/* program is creating its own 3D effect */

	/* 1st - the text box 3D surround */

	/* start outer text box surround with narrow shadow pen */
	orig_pen = SelectObject( disp_dc, nrw_pen_shadow );
	/* start at bottom left */
	MoveToEx( disp_dc, text_rect.left - 2, text_rect.bottom + 1, NULL );
	/* draw bottomn left to top left */
	LineTo( disp_dc, text_rect.left - 2, text_rect.top - 2 );
	/* and top left to top right */
	LineTo( disp_dc, text_rect.right, text_rect.top - 2 );

	/* switch to highlight pen for rest of text box outer surround */
	SelectObject( disp_dc, nrw_pen_light );
	/* draw top right to bottom right */
	LineTo( disp_dc, text_rect.right, text_rect.bottom + 1 );
	/* and bottom right to bottom left */
	LineTo( disp_dc, text_rect.left - 1, text_rect.bottom + 1 );

	/* add inner text box surround with black pen */
	SelectObject( disp_dc, GetStockObject( BLACK_PEN ) );
	/* start at bottom left */
	MoveToEx( disp_dc, text_rect.left - 1, text_rect.bottom, NULL );
	/* draw bottom left to top left */
	LineTo( disp_dc, text_rect.left - 1, text_rect.top - 1 );
	/* and top left to top right */
	LineTo( disp_dc, text_rect.right - 1, text_rect.top - 1 );

	/* switch to shadow pen for rest of inner text box surround */
	SelectObject( disp_dc, nrw_pen_shadow );
	/* draw top right to bottom right */
	LineTo( disp_dc, text_rect.right - 1, text_rect.bottom );
	/* and bottom right to bottom left */
	LineTo( disp_dc, text_rect.left - 1, text_rect.bottom );

#endif

#if	defined( DISPLAY_3D )

	/* 2nd - main window 3D surround */

	/* use the wider light brush for left and top edges of
	   window surround and save original pen */
	SelectObject( disp_dc, wd_pen_light );

	/* go to bottom left of window */
	MoveToEx( disp_dc, win_rect.left + 1, win_rect.bottom, NULL );
	/* draw bottom left to top left */
	LineTo( disp_dc, win_rect.left + 1, win_rect.top );
	/* move all the way to top left */
	MoveToEx( disp_dc, win_rect.left + 1, win_rect.top + 1, NULL );
	/* draw top left to top right */
	LineTo( disp_dc, win_rect.right - 1, win_rect.top + 1 );

	/* use wider shadow pen for right and bottom edges of window surround */
	SelectObject( disp_dc, wd_pen_shadow );
	/* draw top right to bottom right */
	LineTo( disp_dc, win_rect.right - 1, win_rect.bottom - 1 );
	/* and bottom right to bottom left */
	LineTo( disp_dc, win_rect.left + 1, win_rect.bottom - 1 );

	/* 3rd - drop-down list 3D surround - but only if time zones data available */
	if( TRUE == val_tzs )
	{
		/* start drop-down surround with narrow shadow pen */
		SelectObject( disp_dc, nrw_pen_shadow );
		/* start at bottom left */
		MoveToEx( disp_dc, dd_rect.left - 1, dd_rect.bottom, NULL );
		/* draw bottomn left to top left */
		LineTo( disp_dc, dd_rect.left - 1, dd_rect.top - 1 );
		/* and top left to top right */
		LineTo( disp_dc, dd_rect.right, dd_rect.top - 1 );

		/* switch to light pen for rest of drop-down list surround.
		   note that the program does not worry whether or not the
		   drop-down list is visible because it does not matter. */
		SelectObject( disp_dc, nrw_pen_light );
		/* draw top right to bottom right */
		LineTo( disp_dc, dd_rect.right, dd_rect.bottom );
		/* and bottom right to bottom left */
		LineTo( disp_dc, dd_rect.left - 1, dd_rect.bottom );
	}

#endif

#if	defined( DISPLAY_3D ) || defined( TEXTBOX_3D )

	/* put back original pen */
	SelectObject( disp_dc, orig_pen );

#endif

	/* set up co-ordinates for text copy */
	text_out = text_rect;
	/* move to right a little */
	text_out.left += 2;
	/* and down a little */
	text_out.top += 1;

	/* fill text buffer */
	dst_ok = get_buff( obuff, hwnd, yr_offset );
	/* if bad data centre (sic) the text for the error
	   message otherwise expand tabs for normal text,
	   but without prefix (&) processing to avoid, for
	   example, "USA & Canada" being corrupted, and
	   draw text first into memory bitmap. */
	DrawText( mem_dc, obuff, -1, &text_out,
			FALSE == dst_ok ? DT_CENTER : DT_EXPANDTABS | DT_NOPREFIX );
	/* zap result from memory DC to display DC */
	BitBlt( disp_dc, text_rect.left, text_rect.top,
			text_rect.right - text_rect.left - 1, text_rect.bottom - text_rect.top,
			mem_dc, text_rect.left, text_rect.top, SRCCOPY );

	EndPaint( hwnd, &paint_struct );
}


/* return + for numbers greater than zero or empty string so
   that positive -> +x.yy, zero -> 0.00, negative -> -0.00
   as wsprintf will add the - sign, but not the plus. the
   swprintf function would add a + sign but does it for 0.00
   we well as +ve numbers, and I want 0.00 not +0.00 */
static	PWCHAR	pls_blnk( LONG val )
{
	return( val > 0 ? L"+" : L"" );
}


/* it is easy to define our own i_abs() rather than using abs() from the math.h library */
static	INT	i_abs( INT val )
{
	return( val < 0 ? -val : val );
}


/* prepare the DST string */
static	BOOL	get_buff( PWCHAR	obuff_ptr,
			  HWND		hwnd,
			  INT		yr_offset )
{
	WCHAR		std_str[ 256 ];		/* buffer for standard time text */
	LONG		std_tim;		/* LONG value for next standard time */
	WCHAR		sum_str[ 256 ];		/* buffer for summer time text */
	LONG		sum_tim;		/* LONG value for next summer time */
	WCHAR		clk_str[ 256 ];		/* buffer for clock string */
	SYSTEMTIME	loc_time;		/* local time data */
	INT		req_year;		/* year to process for */

	/* one of the following three items can be defined
	   to test for specific characteristics of the data */

#if	defined( TEST_ABSOLUTE )
	/* if this defined then test for arbitrary absolute dates.
	   this will cause an error in Windows NT 3.5 when it tries
	   to do the conversion to Time Zone date and time, because
	   the convertion function does not like absolute dates. */
	tz_info.StandardDate.wDay	= 2;  		/* 2nd .. */
	tz_info.StandardDate.wMonth	= 10;		/* October .. */
	tz_info.StandardDate.wYear	= 1993;		/* .. 1993 */
	tz_info.StandardDate.wDayOfWeek = 6;
	tz_info.DaylightDate.wDay	= 5;  		/* 5th .. */
	tz_info.DaylightDate.wMonth	= 3;		/* March .. */
	tz_info.DaylightDate.wYear	= 1993;		/* .. 1993 */
	tz_info.DaylightDate.wDayOfWeek = 5;
#endif

	GetLocalTime( &loc_time );
	req_year = loc_time.wYear + yr_offset;
	if( TIME_ZONE_ID_UNKNOWN != tz_isdst )
	{
		/* check for range because the program is limited
		   to the years 1901 through to 2099, to avoid any
		   complications caused by 1900 and 2000 not being
		   leap year - what version of Windows will we be
		   up to by 2099? according to Dionysius Exiguus,
		   who suggested numbers for the years, there was
		   no year zero, the calendar going from 1 B.C. to
		   1 A.D., but in reporting here we ignore that! */
		if( (req_year) < 1901 ||
		    (req_year) > 2099 )
		{
			wsprintf( obuff_ptr,
				L"\r\n"
				L"The year %d is out of range\r\n"
				L"\r\n"
				L"Must be from 1901 to 2099\r\n"
				L"\r\n" ,
					req_year );
			/* not need to fiddle with year selection buttons,
			   as they must have been enabled to get here, and
			   that is the state required. */
			return( FALSE );
		}
	}

	/* time zone information is okay - fill data buffers with text
	   and get time values to indicate which to print first */
	std_tim = get_tz_buff( std_str, req_year, L"Standard",
			tz_info.StandardName, tz_info.StandardBias,
			&tz_info.StandardDate, tz_info.Bias, tz_info.DaylightBias );
	sum_tim = get_tz_buff( sum_str, req_year, L"Daylight",
			tz_info.DaylightName, tz_info.DaylightBias,
			&tz_info.DaylightDate, tz_info.Bias, tz_info.StandardBias );

	/* note that in the displays the common elements appear at the
	   same position in the window, so that if the user scrolls
	   through the time zones they do not move about. */

	/* if no timne zone data or no changes are specified */
	if( TIME_ZONE_ID_UNKNOWN == tz_isdst ||
	    ( -1 == std_tim && -1 == sum_tim ) )
	{
		/* no Daylight Saving Time in this time zone - range not relevant.
		   do not ask for report of Standard/Daylight as it does not apply!
		   however, the time zone name is included, even though if the
		   time zone data is available it will already appear in the
		   drop-down list. it helps to fill up the space, and otherwise
		   it would have to be added in when copying to the clipboard.
		   the double \r\n at the end is to keep Word for Win 6 happy. */
		wsprintf( obuff_ptr,
				L"%s\r\n"
				L"\r\n"
				L"Bias from UTC\t=  %s%d%s%02d\r\n"
				L"12%s00 UTC \t=  %02d%s%02d Local Time\r\n"
				L"\r\n"
				L"%s\r\n"
				L"\r\n"
				L"Name\t\t=  %s\r\n"
				L"\r\n",
					tz_times( hwnd, clk_str, FALSE ),
					pls_blnk( tz_info.Bias ), tz_info.Bias / 60, time_sep.string,
								i_abs( tz_info.Bias) % 60,
					time_sep.string,
						( ( 12 * 60 ) - tz_info.Bias ) / 60, time_sep.string,
								i_abs( ( 12 * 60 ) - tz_info.Bias ) % 60,
					tz_disp,
					tz_info.StandardName
						);

		/* grey out year selection buttons */
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_PREV ), FALSE );
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_NOW  ), FALSE );
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_NEXT ), FALSE );

		/* and disallow key selection */
		val_button = FALSE;
		return( TRUE );
	}

	/* now form string to output in a message box. note the date
	   which comes earlier, Standard or Daylight, is shown first.
	   the double \r\n at the end is to keep Word for Win 6 happy. */
	wsprintf( obuff_ptr,
			L"%s\r\n"
			L"\r\n"
			L"Bias from UTC\t=  %s%d%s%02d\r\n"
			L"12%s00 UTC \t=  %02d%s%02d Local Standard Time\r\n"
			L"\r\n"
			L"%s\r\n"
			L"\r\n"
			L"Changes in %d:\r\n"
			L"\r\n"
			L"%s\r\n"
			L"\r\n"
			L"%s\r\n"
			L"\r\n" ,
				tz_times( hwnd, clk_str, TRUE ),
				pls_blnk( tz_info.Bias ), tz_info.Bias / 60, time_sep.string,
							i_abs( tz_info.Bias ) % 60,
				time_sep.string,
					( ( 12 * 60 ) - tz_info.Bias ) / 60, time_sep.string,
							i_abs( ( 12 * 60 ) - tz_info.Bias ) % 60,
				TRUE == val_local ? L"" : tz_disp,
				0 == tz_info.StandardDate.wYear || 0 == tz_info.DaylightDate.wYear ?
							req_year : tz_info.StandardDate.wYear ,
				/* select string for first change in the year,
				   whether to Standard or to Daylight Saving,
				   and then other string for further change. */
				std_tim < sum_tim ? std_str : sum_str,
				std_tim < sum_tim ? sum_str : std_str
					);
	/* set up year selection buttons */
	if( 0 != tz_info.StandardDate.wYear &&
	    0 != tz_info.DaylightDate.wYear )
	{
		/* both years have absolute dates - grey out buttons */
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_PREV ), FALSE );
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_NOW  ), FALSE );
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_NEXT ), FALSE );
		/* and disallow key selection */
		val_button = FALSE;
	}
	else
	{
		/* one or other of the dates is not absolute - enable buttons */
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_PREV ), TRUE );
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_NOW  ), TRUE );
		EnableWindow( GetDlgItem( hwnd, IDC_BUTTON_NEXT ), TRUE );
		/* and allow key selection */
		val_button = TRUE;
	}
	return( TRUE );
}


/* function to return string representing next occurrence of the given
   date, both in terms of the information given and, if Day-In-Month
   format is used, after conversion to a proper date, and fill in the
   text buffer pointed to by the second parameter. return a LONG
   value to indicate which string to output first, to allow for the
   case in the Southern hemisphere where the change to Standard time
   occurs first in the year, before the start of Daylight savings time
   or the Northern hemisphere where the Daylight change occurs first.

   NOTE: that there is a documentation error, as the help file states
   that the changes to and from Daylight Saving Time are defined by a
   SYSTEMTIME structure that gives the date and a UTC time when each
   transition will occur. This is wrong. The time is the time in the
   Local Time Zone, and *NOT* UTC, while the separate 'Bias' field
   gives the offset from UTC of Standard Time in the Local Time Zone.
   If the change really were defined as a UTC time you would have to
   take the 'Bias' into account in determining the time of the change,
   but you don't, so the documentation is wrong. */
static	LONG	get_tz_buff( PWCHAR		obuff_ptr,	/* ptr to text buffer to fill */	     INT		req_year,	/* year to process for */
			     PWCHAR		ch_name_ptr,	/* name of change */
			     PWCHAR		tz_name_ptr,	/* time zone name */
			     LONG		tz_bias,	/* time zone bias */
			     LPSYSTEMTIME	sys_time_ptr,	/* system time to decode */
			     LONG		UTC_bias, 	/* bias from UTC */
			     LONG		alt_bias )	/* other bias */
{
	INT		day_no;				/* the true day in month of the change */
	SYSTEMTIME	time_data;			/* set to true (calendar) date of change */
	WCHAR		date_buff[ 64 ];		/* buffer for date string */
	PWCHAR	auto	date_buff_ptr = date_buff;	/* ptr to date buffer - must be auto */
	WCHAR		time_buff[ 64 ];		/* buffer for time string */
	WCHAR		new_time_buff[ 64 ];		/* buffer for time in zone after change */
	INT		temp_int;			/* temporary integer */

	/* note that there are no checks on the range of the data */

	if( 0 == sys_time_ptr->wMonth )
	{
		/* the OS does not have a change specified */
		wsprintf( obuff_ptr, L"Information Not Available for %s", ch_name_ptr );
		/* return error indicator */
		return( -1 );
	}

	if( 0 != sys_time_ptr->wYear )
	{
		/* this is the easy case, the date is given as
		   absolute so the program just has to returm that. */

		/* convert defined date and time to strings */
		GetTimeFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
			USER_OVERRIDE | TIME_FORCE24HOURFORMAT | TIME_NOTIMEMARKER,
			sys_time_ptr, NULL, time_buff, SIZEOF( time_buff ) );
		GetDateFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
			USER_OVERRIDE | DATE_LONGDATE,
			sys_time_ptr, NULL, date_buff, SIZEOF( date_buff ) );

#if	USER_OVERRIDE != LOCALE_NOUSEROVERRIDE

		/* if the user is allowed to over-ride the date format then for a
		   long date a space is included *if* the day name is omitted from
		   the start, so ... here lose any spaces at the start */
		while( L' ' == *date_buff_ptr )
		{
			date_buff_ptr++;
		}

#endif

		/* alter point to new time after the change - first get new time in minutes -
		   24 hours are added to make sure time stays positive when including the bias */
		temp_int = ( time_data.wHour * 60 ) + time_data.wMinute +
								( 24 * 60 ) - ( tz_bias - alt_bias );
		/* get new hour mod 24 so in range 0 to 23 */
		time_data.wHour = (WORD)( ( temp_int / 60 ) % 24 );
		/* get minutes */
		time_data.wMinute = (WORD)( temp_int % 60 );

		/* convert result to a time string */
		GetTimeFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
			USER_OVERRIDE | TIME_FORCE24HOURFORMAT | TIME_NOTIMEMARKER,
			&time_data, NULL, new_time_buff, SIZEOF( new_time_buff ) );

		/* fill in string buffer */
		wsprintf( obuff_ptr,
			L"     to %s\t=  %ls\r\n"
			L"\tBias\t=  %s%d%s%02d relative to UTC\r\n"
			L"\tDate\t=  %s\r\n"
			L"\tTime\t=  %s  -->  %s   " ,
				ch_name_ptr, tz_name_ptr,
				pls_blnk( tz_bias + UTC_bias ), ( tz_bias + UTC_bias ) / 60,
						time_sep.string, i_abs( tz_bias + UTC_bias ) % 60,
				date_buff_ptr,
				time_buff, new_time_buff );

		/* return relative time for the new point */
		return( relative_time( &time_data ) );
	}
	/* else */

	/* date is in Day-In-Month format, so convert to proper date */

	/* get the day of the week of the 1st of the required month */
	temp_int = Zeller( 1, sys_time_ptr->wMonth, req_year );
	/* use that to find out the date in the month of the first
	   occurrence of the day of the week needed. */
	day_no = ( 7 + sys_time_ptr->wDayOfWeek - temp_int ) % 7 + 1;
	/* find the 'n'th (defined by wDay member) of the particular day */
	day_no += 7 * ( sys_time_ptr->wDay - 1 );
	/* lose a week until the date is within the month, after
	   correcting for February's days in leap years. This does
	   not make corrections for years before 1901 or after 2099! */
	months[ 2 ].days = 28 + ( ( req_year & 3 ) == 0 );
	while( day_no > months[ sys_time_ptr->wMonth ].days )
	{
		/* the MS compiler achieves a quite amazing
		   optimization on this particular loop, by
		   removing altogether an actual loop !! */
		day_no -= 7;
	}
	/* now finally have the normal calendar
	   day in the month for the year required */

	/* set up a SYSTEMTIME data structure with the moment of the
	   time change, updating the year and the true day in month. */
	time_data	= *sys_time_ptr;
	time_data.wYear = (WORD)req_year;
	time_data.wDay  = (WORD)day_no;


	/* convert derived date and time from SYSTEMTIME structure to strings */
	GetTimeFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
			USER_OVERRIDE | TIME_FORCE24HOURFORMAT | TIME_NOTIMEMARKER,
			&time_data, NULL, time_buff, SIZEOF( time_buff ) );
	GetDateFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
			USER_OVERRIDE | DATE_LONGDATE,
			&time_data, NULL, date_buff, SIZEOF( date_buff ) );

#if	USER_OVERRIDE != LOCALE_NOUSEROVERRIDE

	/* if the user is allowed to over-ride the date format then for a
	   long date a space is included *if* the day name is omitted from
	   the start, so ... here lose any spaces at the start */
	while( L' ' == *date_buff_ptr )
	{
		date_buff_ptr++;
	}

#endif

	/* alter point to new time after the change - first get new time in minutes -
	   24 hours are added to make sure time stays positive when including the bias */
	temp_int = ( time_data.wHour * 60 ) + time_data.wMinute + ( 24 * 60 ) - ( tz_bias - alt_bias );
	/* get new hour mod 24 so in range 0 to 23 */
	time_data.wHour = (WORD)( ( temp_int / 60 ) % 24 );
	/* get minutes */
	time_data.wMinute = (WORD)( temp_int % 60 );

	/* convert result to a time string */
	GetTimeFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
			USER_OVERRIDE | TIME_FORCE24HOURFORMAT | TIME_NOTIMEMARKER,
			&time_data, NULL, new_time_buff, SIZEOF( new_time_buff ) );

	/* set up the buffer given with both the Day-In-Month
	   format and with the proper date and the time */
	wsprintf( obuff_ptr,
		L"     to %s\t=  %ls\r\n"
		L"\tBias\t=  %s%d%s%02d relative to UTC\r\n"
		L"\tDate\t=  %s %ls in %ls\r\n"
		L"\t or \t=  %s\r\n"
		L"\tTime\t=  %s  -->  %s" ,
			ch_name_ptr, tz_name_ptr,
			pls_blnk( tz_bias + UTC_bias ), ( tz_bias + UTC_bias ) / 60,
					time_sep.string, i_abs( tz_bias + UTC_bias ) % 60,
			ordinals[ sys_time_ptr->wDay - 1 ],
				days[ sys_time_ptr->wDayOfWeek ].name,
					months[ sys_time_ptr->wMonth ].name,
			date_buff_ptr,
			time_buff, new_time_buff );

	/* return relative time fot the new point */
	return( relative_time( &time_data ) );
}


/* the following piece of magic works out the day of the week, with
   Sunday = 0, using "Zeller's Congruence", which works without
   resort to any conditional code or to any floating point math,
   requiring only simple integer calculation.
   
   the starting year for the domain of this function varies
   depending on where you are in the world, because it only works
   for the Gregorian calendar, and that was not introduced globally
   in the same year. It was first introduced in 1582 in France,
   Italy, Luxembourg, Portugal and Spain, but not till 1584 in most
   of the German Roman Catholic states and parts of the Netherlands.
   However, Denmark and the Dutch and German Protestant states did
   not change till 1699/1700, while Britain did not change till
   1752, Japan till 1873, Egypt till 1875, and Albania, Bulgaria,
   China, Estonia, Latvia, Lithuania, Romania, Turkey and Yugoslavia
   between 1912 and 1917, the Soviet Union in 1918, and Greece in
   1923. Despite all that we assume the Gregorian calendar applies
   globally for the years 1901 and later!
   
   This code was originally in Dr.Dobb's Journal, November 1990,
   Pages 155 and 166 (amended in the edition of February 1991,
   Pages 148 and 149), and was presented in Pascal. The February
   1991 issue defined Zeller's expression as:
   
   	    (m + 1)*26	      K	    J
   	q + ---------- + K + --- + --- - 2*J
		 10	      4	    4

   where the result is taken modulus 7. In this expression q is the
   day in the month, m is the month (but massaged so that January
   and February are months 13 and 14 in the previous year, K is the
   year within a century and J is the century. The later issue notes
   that this is equivalent to:
   
   	    (m + 1)*26	      K	    J
   	q + ---------- + K + --- + --- + 5*J
		 10	      4	    4

   because of the modulus 7, but this avoids the expression ever
   going negative, which would give the wrong result because the
   moduls function on a computer is not the same as that which
   mathematicians (i.e. Zeller) use. This is then equivalent to:

	    (m + 1)*26	 K * 5	  J
   	q + ---------- + ----- + --- + 5*J
		 10	   4	  4

   or:

   	    (m + 1)*26	 (y % 100)*5	 y
   	q + ---------- + ----------- + ----- + 5*(y/100)
		 10	       4       4*100

   where y is the year. but this gives 1 = Sunday, when the program
   wants 0 = Sunday, so:
   
   	      (m + 1)*26   (y % 100)*5     y
   	( q + ---------- + ----------- + ----- + 5*(y/100) + 6 ) % 7
	          10	        4	 4*100
 
   gives what is finally wanted, again adding 6 so that the result
   stays positive, rather than subtracting 1. note that the brackets
   are important in defining the order of evaluation. */
static	INT	Zeller( INT day,
			INT month,
			INT year )
{
	BOOL	correction;

	/* correction is set to 1 for January and February,
	   otherwise 0, and then used to modify year and month */
	year -= correction = ( 14 - month ) / 12;
	month += correction * 12;
	return( ( day +
		( ( ( month + 1 ) * 26 ) / 10 ) +
		( ( ( year % 100 ) * 5 ) / 4 ) +
		( ( year / 100 ) * 5 ) +
		( year / 400 ) +
		6 ) % 7 );
}


/* derive a fiddled time from structure using a 'year' of 12 months each of 31
   days. this is only used for internal comparison and is sufficiently accurate
   for these purposes! only the number of days are needed, since it is unlikely
   that the change to and from Daylight Saving will be in one 24 hour period! */
static	INT	relative_time( LPSYSTEMTIME time_struct )
{
	return( ( ( time_struct->wYear) * 12		/* number of months in the years */
		  + time_struct->wMonth ) * 31		/* number of days in the months */
		  + time_struct->wDay );	        /* number of days altogether */
}


/* function that gets current time for the time zone displayed.
   the system does most of the clever stuff for us - hooray!
   if the calling function supplies a valid buffer pointer the
   space is filled with the clock text, otherwise the clock
   data is written to the caption text. the third parameter
   indicates whether to produce an indication as to whether
   the selected time zone is currently Winter or Summer.
   in this we do not attempt to align the times or dates for
   those time zones that do not use a fixed number of digits. */
static	PWCHAR	tz_times( HWND   hwnd,
			  PWCHAR tz_buff,		/* buffer to fill */
			  BOOL   disp_wint_summ )	/* display winter/summer time */
{
	SYSTEMTIME	sys_time;			/* system time in UTC */
	SYSTEMTIME	zone_time;			/* local time in selected Time Zone */
	SYSTEMTIME	loc_time;			/* local time for this system */
	WCHAR		loc_time_buff[ 64 ];		/* buffer for local time */
	WCHAR		loc_date_buff[ 64 ];		/* buffer for local date */
	WCHAR		utc_time_buff[ 64 ];		/* buffer for UTC time */
	WCHAR		utc_date_buff[ 64 ];		/* buffer for UTC date */
	WCHAR		zone_time_buff[ 64 ];		/* buffer for time in Time Zone */
	WCHAR		zone_date_buff[ 64 ];		/* buffer for date in Time Zone */
	LONG		calc_bias;			/* calculated current bias */
	INT		buff_offset;			/* size of first part of buffer */
	HINSTANCE	han_kern32;			/* handle to Kernel32.dll */
	FARPROC		System_Time_To_Tz_Specific_Local_Time;
							/* function to give time in zone,
							   not present till Windows NT 3.5 */


	/* first check whether the function to convert a UTC time to the local time
	   in a specific time zone is available - it is not on Windows NT 3.1 -
	   here assume that Kernel32.dll is already mapped into the process space! */
	if( TRUE == val_tzs &&
	    NULL != ( han_kern32 = GetModuleHandle( KERNEL_32 ) ) &&
	    NULL != ( System_Time_To_Tz_Specific_Local_Time = GetProcAddress( han_kern32, SYS_TO_TZ ) ) )
	{
		/* Windows NT 3.5 can provide the time in a time zone -
		   start by getting UTC time and the current local time. */
		GetSystemTime( &sys_time );
		/* here we could just

		   GetLocalTime( &loc_time );

		   however, deriving the local time is slightly complicated, as
		   NT 3.5 *SOMETIMES* gets the local time wrong. this is a very
		   minor problem as it only concerns the hour after a change is
		   due, as the system appears not to set the local time properly
		   until an integral number of hours after system boots, but I
		   have not worked out how to reliably demonstrate the problem,
		   as sometimes it seems to work properly. here we don't rely on
		   the GetLocalTime function, but convert the UTC time instead,
		   which always works reliably. this has the extra very minor
		   advantage that the separate calls to the GetSystemTime and
		   GetLocalTime might return different seconds if the one call
		   is just before a seconds change and the other just after.
		   the time in the user selected time zone is derived from UTC
		   (system time). if either year in the Time Zone structure is
		   non-zero, as is allowed for the definition of an absolute
		   date, the conversion function gives an error, but we skip
		   any errors after the first. */
		if( TRUE == System_Time_To_Tz_Specific_Local_Time( &tz_local, &sys_time, &loc_time ) &&
		    TRUE == System_Time_To_Tz_Specific_Local_Time( &tz_info, &sys_time, &zone_time ) )
		{
			/* convert times in UTC, Local and selected Zone */
			GetTimeFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
					USER_OVERRIDE | TIME_FORCE24HOURFORMAT | TIME_NOTIMEMARKER,
					&sys_time, NULL,
					utc_time_buff, SIZEOF( utc_time_buff ) );
			GetTimeFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
					USER_OVERRIDE | TIME_FORCE24HOURFORMAT | TIME_NOTIMEMARKER,
					&loc_time, NULL,
					loc_time_buff, SIZEOF( loc_time_buff ) );
			GetTimeFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
					USER_OVERRIDE | TIME_FORCE24HOURFORMAT | TIME_NOTIMEMARKER,
					&zone_time, NULL,
					zone_time_buff, SIZEOF( zone_time_buff ) );

			if( NULL == tz_buff )
			{
				/* if no response buffer update caption */
				WCHAR	temp_buff[ 256 ];

				/* put times in Windows caption */
				wsprintf( temp_buff, L"UTC=%s Loc=%s Zone=%s",
					utc_time_buff, loc_time_buff, zone_time_buff );
				SetWindowText( hwnd, temp_buff );

				/* and that's it if only caption text needed */
				return( NULL );
			}
			/* else - output to screen */

			/* for window display format the three matching dates */
			GetDateFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
					USER_OVERRIDE | DATE_SHORTDATE,
					&sys_time, NULL,
					utc_date_buff, SIZEOF( utc_date_buff ) );
			GetDateFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
					USER_OVERRIDE | DATE_SHORTDATE,
					&loc_time, NULL,
					loc_date_buff, SIZEOF( loc_date_buff ) );
			GetDateFormatW( MAKELCID( LANG_USED, SORT_DEFAULT ),
					USER_OVERRIDE | DATE_SHORTDATE,
					&zone_time, NULL,
					zone_date_buff, SIZEOF( zone_date_buff ) );

			buff_offset = wsprintf( tz_buff,
					L"UTC       \t=  %s  on  %s\r\n"
					L"Local Time\t=  %s  on  %s\r\n"
					L"Zone Time \t=  %s  on  %s\r\n",
					   	utc_time_buff, utc_date_buff,
						loc_time_buff, loc_date_buff,
						zone_time_buff, zone_date_buff );
			if( TRUE == disp_wint_summ )
			{
				/* to produce an indication of whether it is Winter
				   (Standard) or Summer (Daylight Saving) time first
				   calculate the current bias in the region, by
				   taking the current zone time in minutes away from
				   the current UTC time. 24 hours are added to make
				   sure the bias starts positive, because the two
				   times might be either side of midnight, and then
				   corrected. note that for the local time zone the
				   program does not use the value returned by the
				   GetTimeZoneInformation function to determine if
				   Daylight Saving is in use, as it is as easy to
				   work this out here for all zones as to pick out
				   the local time zone as a special case. */
				calc_bias = ( 24 * 60 ) +
					    ( sys_time.wHour * 60 + sys_time.wMinute ) -
					    ( zone_time.wHour * 60 + zone_time.wMinute );
				/* get result to within the range -12:00 hrs to +11:59 hrs.
				   another case where the MS compiler gets very clever. */
				while( calc_bias >= ( 12 * 60 ) )
				{
					calc_bias -= ( 24 * 60 );
				}
				/* but, that would be wrong for the +12 hours case
				   [not that any Time Zones in the Windows NT 3.5
				   Release Candidate 1 have a bias of +12 hrs and
				   use Daylight Saving Time] so here there is a
				   slight fiddle! if the calculated bias is -12 hrs
				   but the local bias as +12 hrs then the calculated
				   bias is set to + 12hrs, as it is easier to kludge
				   it here than inside the wsprintf() call. */
				if( -( 12 * 60 ) == calc_bias &&
				     ( 12 * 60 ) == tz_info.Bias )
				{
					calc_bias = ( 12 * 60 );
				}
				/* add the indication of whether it is Standard (Winter)
				   Time or Daylight (Summer) Time to the output buffer */
				wsprintf( tz_buff + buff_offset, L"Zone Currently \t=  %s",
							( calc_bias == tz_info.Bias ?
								L"Standard Time (Winter)" :
								L"Daylight Saving (Summer)" ) );
			}
			return( tz_buff );
		}
		else
		{
			/* error from TZ conversion */
			static	BOOL	first = TRUE;
			DWORD		err;
			WCHAR		temp_buff[ 256 ];

			if( first )
			{
				first = FALSE;		/* only report error first time */
				if( ERROR_INVALID_PARAMETER == ( err = GetLastError() ) &&
				    ( 0 != tz_info.StandardDate.wYear ||
				      0 != tz_info.DaylightDate.wYear ) )
				{
					/* if the Time Zone Information has any absolute data,
					   particularly if either year is non-zero, then the
					   SystemTimeToTzSpecificLocalTime() function returns
					   an 'invalid parameter' error (87). report it the
					   first time through and then skip after that. */
					lstrcpy( temp_buff, L"Bad Parameter.\r\n"
						     	    L"\r\n"
							    L"There is an absolute date in\r\n"
							    L"the Time Zone Information./\r\n"
							    L"Cannot determine Time in the Zone." );
				}
				else
				{
					wsprintf( temp_buff, L"Time Zone Conversion Error %lu", err );
				}
				MessageBoxEx( hwnd, temp_buff,
					tzBoxName, MB_OK | MB_ICONEXCLAMATION, LANG_USED );
			}
		}
	}
	/* return pointer to empty string */
	return( L"" );
}


/* function to copy text buffer to clipboard - the box title is added as a header. */
static	VOID	to_clip( HWND hwnd,
			 INT  yr_offset )
{
	HGLOBAL	clip_data;
	PWCHAR	clip_ptr;
	WCHAR	obuff[ 1024 ];

	/* get data to buffer */
	get_buff( obuff, hwnd, yr_offset );
	/* open clipboard and copy data to it */
	if( TRUE == OpenClipboard( hwnd ) )
	{
		if( TRUE == EmptyClipboard() )
		{
			/* allocate global space for data allowing for header and terminator */
			if( NULL != ( clip_data = GlobalAlloc( GMEM_DDESHARE,
							sizeof( WCHAR ) * ( lstrlen( tzCpyRite ) +
										lstrlen( tzBoxName ) +
											lstrlen( obuff ) +
												11 ) ) ) )
			{
				if( NULL != ( clip_ptr = GlobalLock( clip_data ) ) )
				{
					/* first add box title as header */
					lstrcpy( clip_ptr, L"\t" );
					lstrcat( clip_ptr, tzBoxName );
					lstrcat( clip_ptr, L"\r\n\r\n\t" );
					/* then Copyright */
					lstrcat( clip_ptr, tzCpyRite );
					lstrcat( clip_ptr, L"\r\n\r\n" );
					/* and finally the data proper */
	                       		lstrcat( clip_ptr, obuff );
					GlobalUnlock( clip_data );
					/* set clipboard for Unicode text. the clipboard will
					   automagically synthesize the alternatives for ANSI
					   and OEM text so there is no need to create them. */
					if( NULL == SetClipboardData( CF_UNICODETEXT, clip_data ) )
					{
						/* it's not likely, but if the SetClipboardData
						   did not work the memory object is freed here */
						GlobalFree( clip_data );
					}
				}
				else
				{
					/* it's not likely, but if the GlobalLock did not
					   work then the memory object will not be passed
					   to the Clipboard, so it is freed here instead. */
					GlobalFree( clip_data );
				}
			}
		}
		CloseClipboard();
	}
}
